from chunkipy.text_splitters.semantic.sentences.spacy_sentences_text_splitter import SpacySentenceTextSplitter
from chunkipy.text_splitters.semantic.sentences.stanza_sentences_text_splitter import StanzaSentenceTextSplitter

__all__ = [
    "SpacySentenceTextSplitter",
    "StanzaSentenceTextSplitter",
]


import logging
from chunkipy.text_chunker import TextChunker
from chunkipy.text_chunker.data_models import TextPart, Chunk, Chunks, Overlapping


# Configure logging
logging.basicConfig(level=logging.INFO)


__all__ = ["TextChunker",
        "TextPart",
        "Chunk",
        "Chunks",
        "Overlapping"]





from typing import List
from chunkipy.text_splitters.semantic.base_semantic_text_splitter import BaseSemanticTextSplitter
from chunkipy.utils import import_dependencies


class StanzaSentenceTextSplitter(BaseSemanticTextSplitter):
    """Sentence splitter using Stanza for semantic text splitting.
    This class uses Stanza to split text into sentences based on the language detected in the text.
    It supports multiple languages by loading different Stanza models based on the detected language.
    
    Attributes:
        text_limit (int): The maximum length of text to process at once. If None, DEFAULT_LIMIT from base class is applied.
        
    """


    # Mapping from langdetect language codes to Stanza language codes
    # Note: Some languages are not supported by Stanza, so they are set to None
    # Here, the list of languages supported by langdetect (see https://pypi.org/project/langdetect/):   langdetect_languages = ['af','ar','bg','bn','ca','cs','cy','da','de','el','en','es','et','fa','fi','fr','gu','he','hi','hr','hu','id','it','ja','kn','ko','lt','lv','mk','ml','mr','ne','nl','no','pa','pl','pt','ro','ru','sk','sl','so','sq','sv','sw','ta','te','th','tl','tr','uk','ur','vi','zh-cn','zh-tw']
    # Here, the list of languages supported by stanza (invoke stanza.resources.common.list_available_languages()):       stanza_languages = ['af','ar','be','bg','bxr','ca','cop','cs','cu','da','de','el','en','es','et','eu','fa','fi','fr','fro','ga','gd','gl','got','grc','he','hi','hr','hsb','hu','hy','id','it','ja', 'kk','kmr','ko','la','lt','lv','lzh','mr','mt','nl','nn','no','olo','orv','pl','pt','ro','ru','sk','sl','sme','sr','sv','swl','ta','te','tr','ug','uk','ur','vi','wo','zh-hans','zh-hant']
    langdetect_stanza_mapping = {
        'af': 'af',         # Afrikaans
        'ar': 'ar',         # Arabic
        'bg': 'bg',         # Bulgarian
        'bn': None,         # Bengali - not supported in stanza
        'ca': 'ca',         # Catalan
        'cs': 'cs',         # Czech
        'cy': None,         # Welsh - not supported in stanza
        'da': 'da',         # Danish
        'de': 'de',         # German
        'el': 'el',         # Greek
        'en': 'en',         # English
        'es': 'es',         # Spanish
        'et': 'et',         # Estonian
        'fa': 'fa',         # Persian
        'fi': 'fi',         # Finnish
        'fr': 'fr',         # French
        'gu': None,         # Gujarati - not supported in stanza
        'he': 'he',         # Hebrew
        'hi': 'hi',         # Hindi
        'hr': 'hr',         # Croatian
        'hu': 'hu',         # Hungarian
        'id': 'id',         # Indonesian
        'it': 'it',         # Italian
        'ja': 'ja',         # Japanese
        'kn': None,         # Kannada - not supported in stanza
        'ko': 'ko',         # Korean
        'lt': 'lt',         # Lithuanian
        'lv': 'lv',         # Latvian
        'mk': None,         # Macedonian - not supported in stanza
        'ml': None,         # Malayalam - not supported in stanza
        'mr': 'mr',         # Marathi
        'ne': None,         # Nepali - not supported in stanza
        'nl': 'nl',         # Dutch
        'no': 'no',         # Norwegian (Bokmål/Nynorsk combined in langdetect)
        'pa': None,         # Punjabi - not supported in stanza
        'pl': 'pl',         # Polish
        'pt': 'pt',         # Portuguese
        'ro': 'ro',         # Romanian
        'ru': 'ru',         # Russian
        'sk': 'sk',         # Slovak
        'sl': 'sl',         # Slovenian
        'so': None,         # Somali - not supported in stanza
        'sq': None,         # Albanian - not supported in stanza
        'sv': 'sv',         # Swedish
        'sw': None,         # Swahili - not supported in stanza
        'ta': 'ta',         # Tamil
        'te': 'te',         # Telugu
        'th': None,         # Thai - not supported in stanza
        'tl': None,         # Tagalog - not supported in stanza
        'tr': 'tr',         # Turkish
        'uk': 'uk',         # Ukrainian
        'ur': 'ur',         # Urdu
        'vi': 'vi',         # Vietnamese
        'zh-cn': 'zh-hans', # Chinese (Simplified)
        'zh-tw': 'zh-hant'  # Chinese (Traditional)
    }


    def _split(self, text: str) -> List[str]:
        langdetect = import_dependencies(
            extra="spacy-sentence", 
            package_name="langdetect"
        )
        _, DownloadMethod, Pipeline = import_dependencies(
            extra="sentence",
            package_name="stanza",
            attribute_names=["DownloadMethod", "Pipeline"]
        )
        lang = langdetect.detect(text)
        
        stanza_lang = self.langdetect_stanza_mapping.get(lang, None)
        if stanza_lang is None:
            raise ValueError(f"Language '{lang}' is not supported by Stanza for sentence splitting.")
        
        sentence_tokenizer = Pipeline(
            lang=stanza_lang, processors="tokenize", download_method=DownloadMethod.REUSE_RESOURCES
        )
        return [s.text + " " for s in sentence_tokenizer(text).sentences]



import importlib
from typing import List

INSTRUCTIONS = """
chunkipy error: missing {package_name}.

This feature requires additional dependencies:

    $ pip install chunkipy [{extra}]

"""


def format_instructions(*, extra: str, package_name: str) -> str:
    return INSTRUCTIONS.format(extra=extra, package_name=package_name)


class MissingDependencyError(Exception):
    pass


def import_dependencies(extra: str, package_name: str, attribute_names: List[str] = []):
    module = None
    attributes = tuple(None for _ in attribute_names)
    try:
        module = importlib.import_module(package_name)
        attributes = tuple(getattr(module, name) for name in attribute_names)
    except (ImportError, AttributeError) as e:
        raise MissingDependencyError(
            format_instructions(package_name=package_name, extra=extra)
        ) from e
    imports = [module, *attributes] if attributes else module
    return imports



from abc import ABC


class BaseSizeEstimator(ABC):
    """
    Base class for size estimators.
    """

    def estimate_size(self, text: str) -> int:
        """
        Estimate the size of the given text.

        Args:
            text (str): The text to estimate the size of.

        Returns:
            int: The estimated size of the text in bytes.
        """
        raise NotImplementedError("Subclasses must implement the estimate_size method.")


from chunkipy.size_estimators.base_size_estimator import BaseSizeEstimator
from chunkipy.utils import import_dependencies


class OpenAISizeEstimator(BaseSizeEstimator):
    """
    Size estimator that uses OpenAI's tokenization to estimate the size of the text.
    """
    def __init__(self, encoding: str = "cl100k_base"):
        super().__init__()
        tiktoken = import_dependencies(extra="openai-estimator", package_name="tiktoken")
        self.tokenizer = tiktoken.get_encoding(encoding)

    def estimate_size(self, text: str) -> int:
        """
        Estimate the size of the given text using OpenAI's tokenization.

        Args:
            text (str): The text to estimate the size of.

        Returns:
            int: The estimated size of the text in tokens.
        """
        return len(self.tokenizer.encode(text))


from chunkipy.size_estimators.base_size_estimator import BaseSizeEstimator
from chunkipy.size_estimators.word_size_estimator import WordSizeEstimator
from chunkipy.size_estimators.char_size_estimator import CharSizeEstimator
from chunkipy.size_estimators.openai_size_estimator import OpenAISizeEstimator


__all__ = ["BaseSizeEstimator", "WordSizeEstimator", "CharSizeEstimator", "OpenAISizeEstimator"]


from collections import deque
from dataclasses import dataclass, field
from typing import List

@dataclass
class TextPart:
    """Represents a fragment or segment of a complete text, along with its character size.

    :param text: The text of the segment.
    :param size: The size of the text in characters.
    """
    text: str
    size: int


@dataclass  
class Chunk:
    """Represents a single chunk of text, which consists of multiple text parts.
    
    Computed Properties:
    :param text: Represents the full text of the chunk by joining all 'text' values from its 'text parts.
    :param text:parts: A list of TextPart objects that make up the chunk.
    """
    text_parts: List[TextPart] = field(default_factory=list)  # Ensure proper initialization


    @property
    def text(self) -> str:
        """Returns the full concatenated text of the chunk by joining all 'text' values from the TextPart objects.

        Returns:
            str: The full text of the chunk, concatenated from all text parts.
        """
        return ''.join(part.text for part in self.text_parts).strip()


    @property
    def size(self) -> int:
        """Calculates and returns the total size of all TextPart objects within text_parts.

        Returns:
        int: The total size of all TextPart objects.
        """
        return sum(part.size for part in self.text_parts) if self.text_parts else 0
    

    def __repr__(self) -> str:
        return (f"Chunk(text='{self.text}', size={self.size}, "
                f"text_parts={self.text_parts})")


class Chunks(list):
    """A list-like collection of chunks with utility methods for aggregation.

    Inherits from 'list' to act as a standard list, while providing additional methods for aggregated operations.
    """

    def get_all_text_parts(self) -> List[List[str]]:
        """Returns all text parts from each chunk as a list of lists.

        Returns:
            List[List[str]]: A list of lists, where each inner list contains the text parts of a chunk.
        """
        return [chunk.text_parts for chunk in self]


    def get_all_text(self) -> List[str]:
        """Returns the full text from all chunks as a list.
        
         Returns:
            List[str]: A list of strings, where each string is the full text of a chunk.
        """
        return [chunk.text for chunk in self]


class Overlapping(deque):
    """A deque-like collection of TextParts with utility methods for aggregation.
    Inherits from deque to act as a standard deque, while providing additional
    methods for aggregated operations (e.g. size).
    """
    
    @property
    def size(self) -> str:
        """Calculates and returns the total size of all TextPart objects.
        
        Returns:
            int: The total size of all TextPart objects.
        """
        return sum(part.size for part in self) if self else 0

    def __repr__(self) -> str:
        return f"Overlapping(size={self.size}, elements={list(self)})"



from abc import abstractmethod
from typing import List
from chunkipy.text_splitters.base_text_splitter import BaseTextSplitter


class BaseSemanticTextSplitter(BaseTextSplitter):
    """
    Base class for semantic text splitters.
    This class extends BaseTextSplitter and provides a framework for splitting
    text into semantic parts. 

    text_limit attribute helps to control input size for semantic models, that might fail with long texts.
    It is used to limit the size of text processed at once, which is useful for semantic models that may have constraints on input size.
    text_limit does not affect the splitting logic, but rather the size of the text that is passed to the _split method. 
    For example, if your text is 3500 chars and is text_limit is set to 1000, the text will be split into 4 parts of at most 1000 characters before being passed to the _split method.
    
    Args:
        text_limit (int): The maximum length of text to be processed at once.
        If None, defaults to a large value (1,000,000 characters).
    Attributes:
        text_limit (int): The maximum length of text to be processed at once.
        DEFAULT_TEXT_LIMIT (int): Default value for text_limit if not provided.
    Raises:
        NotImplementedError: If the _split method is not implemented in a subclass.
        
    """
    DEFAULT_TEXT_LIMIT = 1000000

    def __init__(self, text_limit: int = None):
        self.text_limit = text_limit or self.DEFAULT_TEXT_LIMIT

    @abstractmethod
    def _split(self, text: str) -> List[str]:
        raise NotImplementedError("Subclasses must implement the _split method.")


    def split(self, text: str) -> List[str]:
        """Split the given text into text parts based on semantic rules.
        This method overrides the split method from BaseTextSplitter and uses the
        _split method to perform the actual splitting. It handles large texts by
        breaking them into smaller chunks based on the text_limit attribute.
        This method ensures that the text is split into manageable parts while
        preserving semantic meaning.
        
        Args:
            text (str): The text to be split.
        Returns:
            List[str]: A list of text parts.
        """
        super()._validate_text(text)
        text_parts = []
        len_last_part = 0

        for i in range(0, len(text), self.text_limit):
            j = i - len_last_part
            partial_text = text[j : i + self.text_limit]
            text_parts.extend(self._split(partial_text))
            if i < (len(text) - self.text_limit):  # if it's not the last iteration
                # Find the index of the last split part within the partial_text, starting from the end.
                # This ensures we account for delimiters, spaces, or any semantic split logic.
                last_part = text_parts.pop() if text_parts else ""
                # Search for the last_part at the end of partial_text
                idx = partial_text.rfind(last_part)
                if idx != -1:
                    # Compute the length of the last part based on its position in partial_text
                    # This ensures that we correctly handle cases where the last part is not at the end of partial_text
                    # (because of spaces, delimiters, or anything else) and we need to adjust the length accordingly.
                    len_last_part = len(partial_text) - idx
                else:
                    # Fallback: use the length of last_part
                    len_last_part = len(last_part)

        return text_parts